#!/usr/bin/env python2

# Restful API

# /pktj/v1/loglevels GET , return loglevel:int <1-8>  OK
# /pktj/v1/loglevels PUT , (loglevel:int <1-8>)       OK
# /pktj/v1/logtypes/<type:int> GET , return enable:true/false
# /pktj/v1/logtypes/<type:int> PUT , (enable:true/false)
# /pktj/v1/lpms/stats/{ipv4|ipv6} GET , return stats
# /pktj/v1/lpms/<ip_addr:ipv4/ipv6> GET , return the lpm entry if found
# /pktj/v1/neighs/{ipv4|ipv6} GET , return neighbors list
# /pktj/v1/acls/{ipv4|ipv6} GET , return acls
# /pktj/v1/acls/{ipv4|ipv6} POST , add new acls
# /pktj/v1/acls/{ipv4|ipv6}/<aclid:int> PUT , modify one acl
# /pktj/v1/acls/{ipv4|ipv6}/<aclid:int> DELETE , delete one acl
# /pktj/v1/stats GET , return current stats
# /pktj/v1/ports/info/<portid:int> GET , return current port info
# /pktj/v1/ports/stats/<portid:int> GET , return current port stats
# /pktj/v1/ports/xstats/<portid:int> GET , return current port extended stats
# /pktj/v1/ports/info/<portid:int> DELETE , clear current port info
# /pktj/v1/ports/stats/<portid:int> DELETE , clear current port stats
# /pktj/v1/ports/xstats/<portid:int> DELETE , clear current port extended stats

from flask import Flask
from flask_restful import reqparse, abort, Api, Resource
import socket
import fcntl
import json
import re
import sys
import ConfigParser
import tempfile
import os
from functools import wraps

app = Flask(__name__)
api = Api(app)

acl4_table = {}
acl6_table = {}
last_acl_file = ''

def build_acl_table(contents):

    acl_table = {}

    acl_id = 0
    for line in iter(contents.splitlines()):
        line = line.lstrip()

        if line[0] == '@': # acl rule
            acl = line[1:]
            acl_params = acl.split()
            if len(acl_params) < 9:
                print 'fetched invalid acl entry'
                continue

            acl = {'src': acl_params[0], 'dst': acl_params[1],
                'src_ports': '%s : %s' % (acl_params[2], acl_params[4]),
                'dst_ports': '%s : %s' % (acl_params[5], acl_params[7]),
                'proto': acl_params[8]}
            acl_table[acl_id] = acl
            acl_id = acl_id + 1

    return acl_table

def load_acl_table(table, ip_version):
    global last_acl_file
    if ip_version not in ('ipv4', 'ipv6'):
        return False

    (file_handle, file_name) = tempfile.mkstemp()
    for id, acl in table.iteritems():
        os.write(file_handle, '@%s %s %s %s %s\n' % (acl['src'], acl['dst'], acl['src_ports'], acl['dst_ports'], acl['proto']))

    pktj_sock.query('acl_add %s %s' % (ip_version, file_name))
    os.close(file_handle)

    # remove old acl file and keep this one
    if len(last_acl_file) > 0:
        os.unlink(last_acl_file)
    last_acl_file = file_name
    return True

class AuthResource(Resource):
    pass
    

class PktjSock(object):
    def __init__(self, so_path, prompt):
        self.so = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.so.connect(so_path)
        flags = fcntl.fcntl(self.so.fileno(), fcntl.F_GETFD)
        flags |= fcntl.FD_CLOEXEC
        fcntl.fcntl(self.so.fileno(), fcntl.F_SETFD, flags)
        self.prompt = prompt

    def query(self, cmd):
        self.so.send("%s\r\n"%cmd)
        data = ''
        while True:
            buf = self.so.recv(8000)
            data += buf
            if len(data) > 20 and data[-6:].startswith(self.prompt):
                break
        print data
        return re.sub(r'%s|%s'%(self.prompt, cmd), '', data)


class LogLevels(AuthResource):
    def __init__(self, **kwargs):
        self.pktj_sock = kwargs['pktj_sock']
        self.parser = reqparse.RequestParser()
        self.parser.add_argument('loglevel')

    def get(self):
        return self.pktj_sock.query('loglevel')

    def put(self):
        args = self.parser.parse_args()
        self.pktj_sock.query('loglevel %s'%args['loglevel'])
        return 202

class LogTypes(AuthResource):
    def __init__(self, **kwargs):
        self.pktj_sock = kwargs['pktj_sock']
        self.parser = reqparse.RequestParser()
        self.parser.add_argument('enable')

    def get(self, logtype):
        return self.pktj_sock.query('logtype %s ' % logtype)

    def put(self, logtype):
        args = self.parser.parse_args()
        self.pktj_sock.query('logtype %s %d' % (logtype, args['enable']))
        return 202

class Acls(AuthResource):
    def __init__(self, **kwargs):
        self.pktj_sock = kwargs['pktj_sock']
        self.parser = reqparse.RequestParser()
        self.parser.add_argument('src_ip', required=True)
        self.parser.add_argument('dst_ip', required=True)
        self.parser.add_argument('src_port_min', required=True)
        self.parser.add_argument('src_port_max', required=True)
        self.parser.add_argument('dst_port_min', required=True)
        self.parser.add_argument('dst_port_max', required=True)
        self.parser.add_argument('proto', required=True)

    def get(self, ipver):
        if ipver == 'ipv4':
            return acl4_table
        elif ipver == 'ipv6':
            return acl6_table
        return 503

    def post(self, ipver):
        if ipver == 'ipv4':
            acl_table = acl4_table
        elif ipver == 'ipv6':
            acl_table = acl6_table
        else:
            return 503

        load_acl_table(acl_table, ipver)

        return 202

    def put(self, ipver):
        if ipver == 'ipv4':
            acl_table = acl4_table
        elif ipver == 'ipv6':
            acl_table = acl6_table
        else:
            return 503

        args = self.parser.parse_args()

        if len(acl_table) == 0:
            acl_id = 0
        else:
            acl_id = max(acl_table.keys(), key=int) + 1
        acl_table[acl_id] = {
            'src': args['src_ip'],
            'dst': args['dst_ip'],
            'src_ports': '%s : %s' % (args['src_port_min'], args['src_port_max']),
            'dst_ports': '%s : %s' % (args['dst_port_min'], args['dst_port_max']),
            'proto': args['proto']
        }

        return 200

class AclsPerId(AuthResource):
    def __init__(self, **kwargs):
        self.pktj_sock = kwargs['pktj_sock']
        self.parser = reqparse.RequestParser()
        self.parser.add_argument('src_ip', required=True)
        self.parser.add_argument('dst_ip', required=True)
        self.parser.add_argument('src_port_min', required=True)
        self.parser.add_argument('src_port_max', required=True)
        self.parser.add_argument('dst_port_min', required=True)
        self.parser.add_argument('dst_port_max', required=True)
        self.parser.add_argument('proto', required=True)

    def put(self, ipver, acl_id):
        if ipver == 'ipv4':
            acl_table = acl4_table
        elif ipver == 'ipv6':
            acl_table = acl6_table
        else:
            return 503

        if acl_id not in acl_table:
            return 502

        args = self.parser.parse_args()

        acl_table[acl_id]['src'] = args['src_ip']
        acl_table[acl_id]['dst'] = args['dst_ip']
        acl_table[acl_id]['src_ports'] = '%s : %s' % (args['src_port_min'], args['src_port_max'])
        acl_table[acl_id]['dst_ports'] = '%s : %s' % (args['dst_port_min'], args['dst_port_max'])
        acl_table[acl_id]['proto'] = args['proto']

        return 200

    def delete(self, ipver, acl_id):
        if ipver == 'ipv4':
            acl_table = acl4_table
        elif ipver == 'ipv6':
            acl_table = acl6_table
        else:
            return 503

        if acl_id not in acl_table:
            return 502

        acl_table.pop(acl_id)

        return 202

pktj_sock = PktjSock('/var/run/pktj.1', 'pktj>')

api.add_resource(LogLevels, '/pktj/v1/loglevels', resource_class_kwargs={'pktj_sock': pktj_sock})
api.add_resource(LogTypes, '/pktj/v1/logtypes/<logtype>', resource_class_kwargs={'pktj_sock': pktj_sock})
api.add_resource(Acls, '/pktj/v1/acls/<ipver>', resource_class_kwargs={'pktj_sock': pktj_sock})
api.add_resource(AclsPerId, '/pktj/v1/acls/<ipver>/<int:acl_id>', resource_class_kwargs={'pktj_sock': pktj_sock})

if __name__ == '__main__':
    if len(sys.argv) != 2:
        print "usage: %s pktj_conf_path" % sys.argv[0]
        sys.exit(1)

    pktj_conf = sys.argv[1]

    config = ConfigParser.ConfigParser()
    config.read(pktj_conf)

    acl4_path = config.get('pktj', 'rule_ipv4')
    acl6_path = config.get('pktj', 'rule_ipv6')

    acl_file = open(acl4_path)
    acl4_contents = acl_file.read()
    acl_file.close()

    acl_file = open(acl6_path)
    acl6_contents = acl_file.read()
    acl_file.close()

    acl4_table = build_acl_table(acl4_contents)
    print 'Bonjour acl4_table: \n%s' % acl4_table

    acl6_table = build_acl_table(acl6_contents)
    print 'Bonjour acl6_table: \n%s' % acl6_table

    app.run(host='0.0.0.0', debug=True)

